/*
 * $Id: ex.c,v 1.26 2002/02/27 16:31:55 cssbz Exp $
 */

/* [Ben] Check for Win32 */
#ifndef __WIN32__
#   if defined(_WIN32) || defined(WIN32) || defined(__CYGWIN__)
#	define __WIN32__
#   endif
#endif

static char rcsid[] = "$Id: ex.c,v 1.26 2002/02/27 16:31:55 cssbz Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <tcl.h>
#include <tk.h>

#include <limits.h>		/* for PATH_MAX */

#include <unistd.h>		/* for exec...() */

#include "../config.h"
#include "../Eden/notation.h"
#include "../Eden/eden.h"
#include "../version.h"

#ifdef DISTRIB
#include <netdb.h>		/* for MAXHOSTNAMELEN on Solaris [Ash] */
#include <sys/param.h>		/* for MAXHOSTNAMELEN on Linux [Ash] */
#endif

#include "../Eden/emalloc.h"
#ifdef HAVE_DMALLOC
#include <dmalloc.h>
#endif

extern int  run(short, void *, char *);
extern void queue(char *, char *);
extern void warningf(char *, ...);
extern void pushMasterStack(char *);
extern void popMasterStack(void);
extern void evaluate(char *, char *);
extern void terminate(int);

extern int  eden_interrupt_handler(ClientData, Tcl_Interp *, int, char *[]);
extern int  dumpeden(ClientData, Tcl_Interp *, int, char *[]);
extern int  dumpscout(ClientData, Tcl_Interp *, int, char *[]);
extern int  dumpdonald(ClientData, Tcl_Interp *, int, char *[]);
extern int  viewOptions(ClientData, Tcl_Interp *, int, char *[]);

#ifdef DISTRIB
extern int  dumpLSD(ClientData, Tcl_Interp *, int, char *[]);
extern int  getNotation(ClientData, Tcl_Interp *, int, char *[]);  /* sun */
#endif /* DISTRIB */

#include <tk.h>
/* Tcl caches environment variables: so we must use Tcl's versions of
   the getenv and putenv functions [Ash] */
/* Tcl 8.2.0 Makefile implies don't use the Tcl getenv [Ash] */
#define putenv Tcl_PutEnv
#define setenv TclSetEnv
#define unsetenv TclUnsetEnv
#include <tcl.h>

Tcl_Interp *interp;
Tk_Window   mainWindow;

static int  edenCmd(ClientData, Tcl_Interp *, int, char *[]);
static int  todoCmd(ClientData, Tcl_Interp *, int, char *[]);
static int  evaluateCmd(ClientData, Tcl_Interp *, int, char *[]);
static int  quitCmd(ClientData, Tcl_Interp *, int, char *[]);
static int  refreshCmd(ClientData, Tcl_Interp *, int, char *[]);
static int  gotoCmd(ClientData, Tcl_Interp *, int, char *[]);
int         Tcl_AppInit(Tcl_Interp *);
void        EXinit(int, char **);

extern char *ex_sptr, ex_sbuf[];
char       *sptr;
extern char *progname;
extern char *libLocation;

#define STRING_DEV      0

/* This implements the Tcl "eden" command [Ash] */
static int
edenCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    notationType notation;

    if (argc != 2) {
	warningf("Tcl: eden: wrong # of arguments - expecting 1, got %d",
		argc - 1);
    }
    notation = currentNotation;
    changeNotation(EDEN);
    pushMasterStack("user");
    run(STRING_DEV, argv[1], 0);
    popMasterStack();
    changeNotation(notation);
    return TCL_OK;
}

static int
todoCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    if (argc != 2) {
	warningf("Tcl: eden: wrong # of arguments - expecting 1, got %d",
		 argc - 1);
    }
    queue(argv[1], "interface");
    return TCL_OK;
}

static int
evaluateCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    if (argc != 2) {
	warningf("Tcl: evaluate: wrong # of arguments - expecting 1, got %d",
		argc - 1);
    }
    evaluate(argv[1], "input");
    return TCL_OK;
}

static int
quitCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    terminate(0);
    return TCL_ERROR;
}

static int
restartCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    extern int startupargc;
    extern char **startupargv;
    extern char *startupcwd;
    pid_t pid;

/* 
    fprintf(stderr, "startupargc = %d\n", startupargc);
    fprintf(stderr, "startupargv[0] = %s\n", startupargv[0]);
    fprintf(stderr, "startupargv[1] = %s\n", startupargv[1]);
 */

    /* Need to restore the initial working directory as it is changed
       whilst EDEN runs by include()'ing models etc */
    chdir(startupcwd);
    
    /* Seems can't just exec the EDEN binary again, perhaps because
       file descriptors aren't closed?  So fork a new EDEN, and stop
       this one. */
    switch (pid = fork()) {
      case -1:
        perror("can't fork");
        exit(-1);
      case 0:
        /* this process is the newly forked process -- start up again
           with the original arguments */
	fprintf(stderr, "starting new EDEN process...\n");
        if (execvp(startupargv[0], startupargv) == -1) {
          perror("can't exec");
          exit(-1);
        }
      default:
        /* this process is the old process -- just stop */
        fprintf(stderr, "stopping old EDEN process...\n");
        exit(0);
    }

    /* should never get here */
    return TCL_ERROR;
}

static int
refreshCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
    Tk_Window   tkwin;
    XEvent      e;

    if (argc != 2) {
	warningf("Tcl: refresh: wrong # of arguments - expecting 1, got %d",
		 argc - 1);
    }
    tkwin = Tk_NameToWindow(interp, argv[1], mainWindow);
    e.xexpose.type = Expose;
    e.xexpose.send_event = True;
    e.xexpose.display = Tk_Display(tkwin);
    e.xexpose.window = Tk_WindowId(tkwin);
    e.xexpose.x = 0;
    e.xexpose.y = 0;
    e.xexpose.width = Tk_Width(tkwin);
    e.xexpose.height = Tk_Height(tkwin);
    e.xexpose.count = 1;
    XSendEvent(Tk_Display(tkwin), Tk_WindowId(tkwin), True, NoEventMask, &e);
    return TCL_OK;
}

static int
gotoCmd(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
  /* I guess removed from Windows by [Ben] due to different
     interaction style there... although I'm not even sure that this
     function is used at all. [Ash] */
#ifndef __WIN32__
    Tk_Window   tkwin;

    if (argc != 4) {
	interp->result = "wrong # args";
	return TCL_ERROR;
    }
    if (strlen(argv[1]) == 0) {
	XWarpPointer(Tk_Display(mainWindow), 0, None,
		     0, 0, 0, 0, atoi(argv[2]), atoi(argv[3]));
    } else {
	tkwin = Tk_NameToWindow(interp, argv[1], mainWindow);
	XWarpPointer(Tk_Display(tkwin), 0, Tk_WindowId(tkwin),
		     0, 0, 0, 0, atoi(argv[2]), atoi(argv[3]));
    }
    interp->result = "";
    return TCL_OK;
#endif
}

/* Tcl_Eval with Error Check - uses Tcl_EvalEx for speed if we have it
   [Ash] */
#ifdef HAVE_TCL_EVALEX
#define TCL_EVALEX(i, s, n, f) Tcl_EvalEx(i, s, n, f)
#else
#define TCL_EVALEX(i, s, n, f) Tcl_Eval(i, s)
#endif

inline void Tcl_EvalEC(Tcl_Interp * interp, char *script) {
#ifdef DEBUG
  if (Debug & 128)
    fprintf(stderr, "Tcl_EvalEC(\"%s\")\n", script);
#endif
  if (TCL_EVALEX(interp, script, -1, 0) != TCL_OK) {
    errorf("Tcl error (Tcl_EvalEC): %s", Tcl_GetStringResult(interp));
  }
}

/* Tcl_GlobalEval with Error Check - uses Tcl_EvalEx for speed [Ash] */
inline void Tcl_GlobalEvalEC(Tcl_Interp * interp, char *script) {
#ifdef DEBUG
  if (Debug & 128)
    fprintf(stderr, "Tcl_GlobalEvalEC(\"%s\")\n", script);
#endif
  if (TCL_EVALEX(interp, script, -1, TCL_EVAL_GLOBAL) != TCL_OK) {
    errorf("Tcl error (Tcl_GlobalEvalEC): %s", Tcl_GetStringResult(interp));
  }
}

#ifdef __CYGWIN__
int cygwin_conv_to_full_win32_path_tcl(ClientData clientData,
				       Tcl_Interp *interp,
				       int argc, char *argv[]) {
  extern cygwin_conv_to_full_win32_path(const char *path, char *win32_path);
  char winpath[PATH_MAX];

  if (argc != 2) {
    warningf("cygwin_conv_to_full_win32_path: requires 1 argument, got %d",
	     argc - 1);
  }

  if (!cygwin_conv_to_full_win32_path(argv[1], winpath)) {
    Tcl_SetResult(interp, winpath, TCL_STATIC);
    return TCL_OK;
  } else {
    return TCL_ERROR;
  }
}
#endif

int
Tcl_AppInit(Tcl_Interp * interp)
{
    if (Tcl_Init(interp) == TCL_ERROR)
	return TCL_ERROR;
    if (Tk_Init(interp) == TCL_ERROR)
	return TCL_ERROR;

#ifdef __APPLE__
    TkMacOSXInitAppleEvents(interp);
    // The next line is not required with Mac OS X 10.2.4, Tcl/Tk 8.4.2
    //    TkMacOSXInitMenus(interp);
#endif

    Tcl_CreateCommand(interp, "quit", quitCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "evaluate", evaluateCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "eden", edenCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "todo", todoCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "refresh", refreshCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "goto", gotoCmd, NULL, NULL);
    Tcl_CreateCommand(interp, "dumpeden", dumpeden, NULL, NULL);
    Tcl_CreateCommand(interp, "dumpscout", dumpscout, NULL, NULL);
    Tcl_CreateCommand(interp, "dumpdonald", dumpdonald, NULL, NULL);
    Tcl_CreateCommand(interp, "setupViewOptions", viewOptions, NULL, NULL);
    Tcl_CreateCommand(interp, "interrupt", eden_interrupt_handler, NULL,
		      NULL);
    Tcl_CreateCommand(interp, "restart", restartCmd, NULL, NULL);
#ifdef DISTRIB
    Tcl_CreateCommand(interp, "dumpLSD", dumpLSD, NULL, NULL);
    Tcl_CreateCommand(interp, "getOldNotation", getNotation, NULL, NULL);
    Tcl_CreateCommand(interp, "getCurrentNotation", getNotation, NULL, NULL);
#endif /* DISTRIB */

#ifdef __WIN32__
    Tcl_CreateCommand(interp, "cygwin_conv_to_full_win32_path", 
		      cygwin_conv_to_full_win32_path_tcl, NULL, NULL);
#endif

#ifdef FAILED_CONSOLE_ATTEMPT
    /* This doesn't work: tcl output via puts etc goes to this console
       window (and Tcl commands typed there are executed!), but output
       from Eden to stdout etc does not */
#ifdef __APPLE__
    /*
     * If we don't have a TTY, then use the Tk based console
     * interpreter instead.
     */

    if (ttyname(0) == NULL) {
      Tk_InitConsoleChannels(interp);
      /*      Tcl_RegisterChannel(interp, Tcl_GetStdChannel(TCL_STDIN));
      Tcl_RegisterChannel(interp, Tcl_GetStdChannel(TCL_STDOUT));
      Tcl_RegisterChannel(interp, Tcl_GetStdChannel(TCL_STDERR));
      */
      Tcl_RegisterChannel(interp, Tcl_MakeFileChannel(stdin, TCL_READABLE));
      Tcl_RegisterChannel(interp, Tcl_MakeFileChannel(stdout, TCL_WRITABLE));
      Tcl_RegisterChannel(interp, Tcl_MakeFileChannel(stderr, TCL_WRITABLE));

      if (Tk_CreateConsoleWindow(interp) == TCL_ERROR) {
	return TCL_ERROR;
      }
      Tcl_Eval(interp, "console show");
    }
#endif
#endif /* FAILED_CONSOLE_ATTEMPT */

    return TCL_OK;
}


void
EXinitTcl(int argc, char *argv[])
{
    interp = Tcl_CreateInterp();

    /* The following is copied from
       /dcs/emp/empublic/src/tk8.3.3/generic/tkMain.c
    */

    Tcl_FindExecutable(argv[0]);

    /*
     * Set the "tcl_interactive" variable.
     */

    Tcl_SetVar(interp, "tcl_interactive", "1", TCL_GLOBAL_ONLY);


    /* Tcl_SetVar(interp, "argv0", argv[0], TCL_GLOBAL_ONLY); */

    /*
     * Invoke application-specific initialization.
     */
    if (Tcl_AppInit(interp) != TCL_OK) {
      fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
      exit(1);
    }


    /* Set the "tcl_interactive" variable. */
    //Tcl_SetVar(interp, "tcl_interactive", "1", TCL_GLOBAL_ONLY);

    /* Invoke application-specific initialization. */
    //if (Tcl_AppInit(interp) != TCL_OK) {
    //fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
    //exit(1);
    //    }
}

void
EXinit(int argc, char *argv[])
{
    extern char *progname;
#ifdef DISTRIB
    extern char hostName[MAXHOSTNAMELEN]; /* provided in Eden/main.c [Ash] */
    extern int socketPort;  /* for distributed tkEden --sun */
    extern char *current_Notation;
    extern char *hostNamePtr;
    extern Int *synchronize;
    extern int isServer;
    extern char *loginnamePtr;
#endif /* DISTRIB */
    char       *name = "/edenio.tcl";
    extern char currentAgentName[128];
    extern char currentAgentType[3];
    extern char *currentAgentNamePtr;
    extern char *currentAgentTypePtr;
    char        fullname[255];
    /* This seems safe to do: if $DISPLAY is unset, Tcl_AppInit will
       have failed earlier [Ash] */
    char       *display = getenv("DISPLAY");
    char       *class = (char *) ckalloc((unsigned) (strlen(progname) + 1));
    char *envcontents;
    /* These must be static so they don't go out of scope and
       disappear, leaving Tcl saying "where are they", causing Seg
       faults etc.  D'oh!  [Ash] */
    static char *version;		/* version of tkeden [Ash] */
    static char *svnversion;	/* Subversion revision [Ash] */
    static char *variant;		/* whether tty, tk, d, ? [Ash] */
    static char *webSite;		/* address of the EM web site [Ash] */
    static char *win32version;	/* Win32 version number [Ben] */
    static char *sasamiAvail;	/* whether we have sasami [Ash] */
    static char *apple;		/* whether we are running on an Apple */

    /* EXinitTcl(argc, argv); */

    strcpy(class, progname);
    class[0] = toupper((unsigned char) class[0]);

    /* EXinitTcl() should have already been called by main.c [Ash] */

    mainWindow = Tk_MainWindow(interp);
    ckfree(class);
    if (mainWindow == NULL) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }

    /* Work out file name of Tcl interface code */
    strcpy(fullname, libLocation);
    strcat(fullname, name);

    currentAgentNamePtr = currentAgentName;
    currentAgentTypePtr = currentAgentType;

#ifdef DISTRIB
    if (isServer) {
      /* DISTRIB, server */
       Tcl_LinkVar(interp, "svcPort", (char *) &socketPort, TCL_LINK_INT);
       Tcl_LinkVar(interp, "eshost", (char *) &hostNamePtr, TCL_LINK_STRING);
       Tcl_LinkVar(interp, "oldNotation", (char *) &current_Notation,
		   TCL_LINK_STRING);
       Tcl_LinkVar(interp, "oldAgentType", (char *) &currentAgentTypePtr,
		   TCL_LINK_STRING);
       Tcl_LinkVar(interp, "oldAgentName", (char *) &currentAgentNamePtr,
		   TCL_LINK_STRING);
    } else {
      /* DISTRIB, client */
       Tcl_LinkVar(interp, "esport", (char *) &socketPort, TCL_LINK_INT);
       Tcl_LinkVar(interp, "eshost", (char *) &hostNamePtr, TCL_LINK_STRING);
       Tcl_LinkVar(interp, "currentNotation", (char *) &current_Notation,
		   TCL_LINK_STRING);
       Tcl_LinkVar(interp, "dtkedenloginname", (char *) &loginnamePtr,
           TCL_LINK_STRING);
    }
    /* printf("sockPort = %i", socketPort); */
    Tcl_LinkVar(interp, "synchronize", (char *) synchronize, TCL_LINK_INT);

    Tcl_LinkVar(interp, "_dtkeden_isServer", (char *) &isServer,
		TCL_LINK_BOOLEAN | TCL_LINK_READ_ONLY);
#endif /* DISTRIB */

    /* Create Tcl variables... */
    version = Tcl_Alloc(TKEDEN_VERSION_MAX_LEN);
    *version = 0;
    strcat(version, TKEDEN_VERSION);
    Tcl_LinkVar(interp, "_tkeden_version", (char *) &version,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    svnversion = Tcl_Alloc(50); /* assume svnversion string will never be huge */
    *svnversion = 0;
    strcat(svnversion, EDEN_SVNVERSION);
    Tcl_LinkVar(interp, "_eden_svnversion", (char *) &svnversion,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

   variant = Tcl_Alloc(TKEDEN_VARIANT_MAX_LEN);
    *variant = 0;
    strcat(variant, TKEDEN_VARIANT);
    Tcl_LinkVar(interp, "_tkeden_variant", (char *) &variant,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    webSite = Tcl_Alloc(TKEDEN_WEB_SITE_MAX_LEN);
    *webSite = 0;
    strcat(webSite, TKEDEN_WEB_SITE);
    Tcl_LinkVar(interp, "_tkeden_web_site", (char *) &webSite,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    sasamiAvail = Tcl_Alloc(2);
    *sasamiAvail = 0;
#ifdef WANT_SASAMI
    strcat(sasamiAvail, "1");
#else
    strcat(sasamiAvail, "0");
#endif
    Tcl_LinkVar(interp, "_tkeden_sasamiAvail", (char *) &sasamiAvail,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    /* [Ben] Create _tkeden_win32_version variable (used in TCL to
             check for Win32)
             _tkeden_win32_version is "0.0" if not running on Win32
             This is set in version.h */

    win32version = Tcl_Alloc(TKEDEN_WIN32_VERSION_MAX_LEN);
    *win32version = 0;
    strcat(win32version, TKEDEN_WIN32_VERSION);
    Tcl_LinkVar(interp, "_tkeden_win32_version", (char *) &win32version,
	 	TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    /* Hack something so that Tcl can tell when we are running on an
       Apple, and do the Apple menu etc */
    apple = Tcl_Alloc(5);
    *apple = 0;
#ifdef __APPLE__
    strcat(apple, "1");
#else
    strcat(apple, "0");
#endif
    Tcl_LinkVar(interp, "_tkeden_apple", (char *) &apple,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);

    if (Tcl_EvalFile(interp, fullname) != TCL_OK) {
      fprintf(stderr, "%s: Tcl error whilst evaluating %s\n", progname,
	      fullname);
      fprintf(stderr, "Tcl errorInfo (from EXinit): %s\n",
	      Tcl_GetVar(interp, "errorInfo", 0));
      exit(1);
    }

#ifdef DISTRIB
/*     setSocketPort(); */
#endif /* DISTRIB */

}
