/*
 * $Id: runset.c,v 1.9 2001/12/06 22:48:41 cssbz Exp $
 */

static char rcsid[] = "$Id: runset.c,v 1.9 2001/12/06 22:48:41 cssbz Exp $";

#include <stdlib.h>
#include <string.h>

#include "../config.h"
#include "runset.h"
#include "eden.h"

#include "emalloc.h"
#ifdef HAVE_DMALLOC
#include <dmalloc.h>
#endif

void
initRunSet(RunSet * rs)
{
    rs->nitems = 0;
    rs->head = rs->tail = (Action *) emalloc(sizeof(Action));
    rs->tail->s = 0;
    rs->tail->master = 0;
    rs->tail->next = 0;
}

void
clearRunSet(RunSet * rs)
{
    while (rs->nitems)
	rmAction(rs);
}

Action
getAction(RunSet * rs)
{
    Action     *ptr = rs->head->next;
    Action      a;

    a.s = ptr->s;
    a.master = ptr->master;
    rs->head->next = rs->head->next->next;
    free(ptr);
    rs->nitems--;
    if (rs->nitems == 0)
	rs->tail = rs->head;
    return a;
}

void
rmAction(RunSet * rs)
{
    Action     *ptr = rs->head->next;

    free(ptr->s);
    rs->head->next = rs->head->next->next;
    free(ptr);
    rs->nitems--;
    if (rs->nitems == 0)
	rs->tail = rs->head;
}

void
addAction(RunSet * rs, char *s, char *master)
{
    rs->tail->next = (Action *) malloc(sizeof(Action));
    rs->tail = rs->tail->next;
    rs->tail->s = strdup(s);
    rs->tail->master = master;
    rs->tail->next = 0;
    rs->nitems++;
}

#ifdef DEBUG
/* for debugging purpose */

#include <stdio.h>

void
printRunSet(RunSet * rs)
{
    Action     *ptr;
    int n = 0;

    if (rs->nitems == 0) {
      fprintf(stderr, "** printRunSet: no items in RS\n");
    } else {
      for (ptr = rs->head->next; ptr; ptr = ptr->next) {
	fprintf(stderr, "** printRunSet: %d: %s\n", ++n, ptr->s);
      }
    }
}

#endif
