/*
 * $Id: tree.h,v 1.5 2001/07/27 17:04:34 cssbz Exp $
 */

/************************************************************************
 *									*
 *   int_tree(int i) - build a tree which denotes an integer		*
 *									*
 *   str_tree(char *s) - build a tree which denotes a string constant	*
 *									*
 *   def_tree(symbol *s, tree *def) - build a parse tree of definition	*
 *				      variable s is defined to be `def'	*
 *									*
 *   sym_tree(symbol* sym, int type) - build a tree which denotes a	*
 *				      variable of type `type'		*
 *									*
 *   tree2(tree *l, int opcode) - build a tree with only its left son	*
 *			(= `l') and opcode(= `opcode') are defined	*
 *									*
 *   tree3(tree *l, int opcode, tree*r) - build a tree with left son	*
 *			= `l', opcode = `opcode' and right son = `r'	*
 *									*
 *   freetree(tree *t) - free the tree `t'				*
 *									*
 ************************************************************************/

typedef struct {
    int         type;
    struct tree *frame;
    struct tree *string;
    struct tree *box;
    struct tree *pict;
    struct tree *xmin;
    struct tree *ymin;
    struct tree *xmax;
    struct tree *ymax;
    struct tree *font;
    struct tree *bgcolor;
    struct tree *fgcolor;
    struct tree *bdcolor;
    struct tree *border;
    struct tree *bdtype;
    int         align;
  //    int         sensitive;
    struct tree *sensitive;
}           WinStruct;

typedef struct {
    int         change;
    union {
	int         type;
	struct tree *frame;
	struct tree *string;
	struct tree *box;
	struct tree *pict;
	struct tree *xmin;
	struct tree *ymin;
	struct tree *xmax;
	struct tree *ymax;
	struct tree *font;
	struct tree *bgcolor;
	struct tree *fgcolor;
	struct tree *bdcolor;
	struct tree *border;
	struct tree *bdtype;
	int         align;
      //	int         sensitive;
        struct tree *sensitive;
    }           f;		/* window attributes */
}           WinField;

struct tree {
    int         op;
    union {
	char       *s;
	char        c;
	double      i;
	struct symbol *v;
	struct tree *t;
	WinStruct   w;
    }           l, r;
};
typedef struct tree tree;

extern tree *int_tree();
extern tree *inthonest_tree();
extern tree *str_tree();
extern tree *win_tree();
extern tree *def_tree();
extern tree *sym_tree();
extern tree *tree2();
extern tree *tree3();
extern void freetree();
